'''
  Folding@Home Client Control (FAHControl)
  Copyright (C) 2010-2011 Stanford University

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import threading
import traceback
import collections
import re
import shlex

from fah import *
from fah.util import make_row
from fah.util import OrderedDict

debug = True
debug = False

def parse_lines(data):
    if not data: return

    for line in re.findall(r'([^\r\n]+)\s*', data):
        line = line.strip()
        if line: yield line


def parse_options(data):
    options = {}

    for line in parse_lines(data):
        name, value = line.split(':', 1)

        # Filter out <invalid>
        if value == '<invalid>':
            print 'Warning:', line
            value = ''

        options[name] = value.strip().decode('string_escape')

    return options


class ClientUpdater:
    connect_rate = 5
    update_rate = 5

    def __init__(self, app, address, port, password, follow_log):
        # Lists
        self.option_names = app.client_option_widgets.keys()

        # Log
        self.log_open = False
        self.log_queue = collections.deque()

        # Commands
        self.connect_fail_reason = ''
        self.command_queue = collections.deque()
        self.failed_commands = 0

        # Timers
        self.last_connection = 0
        self.last_update_attempt = 0

        # Events
        self.shutdown = threading.Event()
        self.online = threading.Event()
        self.connecting = threading.Event()
        self.selected = threading.Event()
        self.updated = threading.Event()
        self.status_updated = threading.Event()
        self.follow_log = threading.Event()
        if follow_log: self.follow_log.set()
        self.refresh_log = threading.Event()
        self.clear_log = threading.Event()

        # Objects
        self.config = ClientConfig()
        self.com = ClientCom(self.shutdown, address, port, password)

        # Thread
        self.lock = threading.RLock()
        self.commands_flushed = threading.Condition(self.lock)
        self.thread = threading.Thread(target = self.run)
        self.thread.setDaemon(True)
	self.thread.start()


    # External interface
    def get_connect_fail_reason(self):
        try:
            self.lock.acquire()
            return self.connect_fail_reason
        finally: self.lock.release()
        

    def get_status(self):
        if self.online.isSet():
            if self.updated.isSet(): return 'Online'
            if self.selected.isSet(): return 'Updating'
            return 'Online'
        else:
            if self.connecting.isSet(): return 'Connecting'
            return 'Offline'


    def get_config(self):
        try:
            self.lock.acquire()
            return copy.deepcopy(self.config)
        finally: self.lock.release()


    def get_status_config(self):
        try:
            self.lock.acquire()
            queue = copy.deepcopy(self.config.queue)
            slot_info = copy.deepcopy(self.config.slot_info)
            return queue, slot_info
        finally: self.lock.release()


    def clear_errors(self):
        try:
            self.lock.acquire()
            self.failed_commands = 0
        finally: self.lock.release()


    def queue_command(self, cmd):
        try:
            self.lock.acquire()
            self.command_queue.appendleft(cmd)
        finally: self.lock.release()


    def flush_commands(self):
        try:
            self.lock.acquire()
            while len(self.command_queue):
                self.commands_flushed.wait()

            return self.failed_commands
                
        finally:
            self.lock.release()
            

    def get_log_updates(self):
        try:
            self.lock.acquire()
            
            updates = []
            while len(self.log_queue):
                updates.append(self.log_queue.pop())

            return updates

        finally: self.lock.release()


    def close(self):
        self.shutdown.set()
        self.thread.join()
        self.com.close()


    # Com
    def connect(self):
        try:
            self.lock.release()
            return self.com.connect()
        finally:
            self.lock.acquire()


    def send_command(self, cmd):
        try:
            self.lock.release()
            return self.com.send_command(cmd)
        finally:
            self.lock.acquire()


    def send(self, data):
        try:
            self.lock.release()
            return self.com.send(data)
        finally:
            self.lock.acquire()


    def receive_block(self):
        try:
            self.lock.release()
            return self.com.receive_block()
        finally:
            self.lock.acquire()
        

    # Log functions
    def log_add(self, data):
        if not data: return

        # Remove color codes
        data = re.sub(r'\033\[\d\d?m', '', data)

        self.config.log.append(data)
        self.log_queue.appendleft(data)


    def log_download(self):
        try:
            self.log_clear()

            # Open the log
            self.send_command('log open')

            more = True
            while more:
                # Get the data
                self.send('log read\n')
                more, data = self.receive_block()

                self.log_add(data)

            if self.follow_log.isSet():
                self.log_open = True

        except:
            if debug: traceback.print_exc()


    def log_update(self):
        # Clear
        if self.clear_log.isSet():
            self.clear_log.clear()
            self.log_clear()

        # Refresh
        if self.refresh_log.isSet():
            try:
                self.log_download()
                self.refresh_log.clear()
            except:
                if debug: traceback.print_exc()
            return

        # Follow
        if not self.follow_log.isSet(): return

        if not self.log_open:
            # Open the log
            self.send_command('log tail')
            self.log_open = True

        # Get the data
        if debug: print 'Updating log'
        self.send('log read\n')
        more, data = self.receive_block()

        self.log_add(data)


    def log_clear(self):
        self.config.log = []
        self.log_queue.clear()
        self.log_queue.appendleft('') # Signal GUI to clear the log

    
    # Download functions
    def download_queue_info(self):
        # Get the data
        data = self.send_command('queue-info')

        # Parse it
        queue = []
        for line in parse_lines(data):
            keys = {}
            for pair in shlex.split(line):
                key, value = pair.split(':', 1)
                keys[key.lower()] = value.strip().replace('_', ' ')

            queue.append(keys)

        return queue


    def download_info(self):
        data = self.send_command('info').strip()

        # Parse it
        info = OrderedDict()
        section = None

        for line in parse_lines(data):
            line = line.strip()
            if not line: continue

            if line[0] == '*':
                section = OrderedDict()
                m = re.match(r'^\*+ ([^*]+) \*+$', line)
                if m:
                    name = m.group(1)
                    info[name] = section

            elif section is not None:
                try:
                    name, value = line.split(':', 1)
                    section[name] = value.strip()
                except:
                    print 'Failed to parse info info line: ' + line

        return info


    def download_options(self):
        # Get option names
        names = self.option_names
        names = map(lambda name: name.replace('_', '-'), names)

        # Get option data
        data = self.send_command('options %s *' % ' '.join(names))
        
        # Parse it
        return parse_options(data)


    def download_slot_options(self, id):
        # Get the data
        data = self.send_command('slot-options %d' % id)

        # Parse it
        return parse_options(data)


    def download_slot_info(self):
        # Get the data
        data = self.send_command('slot-info')

        # Parse it
        slot_info = []
        for line in parse_lines(data):
            id, rest = line.split(':', 1)
            id = int(id)
            status, description = rest.strip().split(' ', 1)

            slot_info.append((id, status, description.strip()))

        return slot_info


    def download_slots(self):
        slot_info = self.download_slot_info()
        
        slots = []
        for id, status, description in slot_info:
            options = self.download_slot_options(id)
            slot = SlotConfig(id, status, description, options)
            slots.append(slot)

        return slots


    def update_client_config(self):
        self.config.options = self.download_options()
        self.config.slots = self.download_slots()
        self.config.info = self.download_info()

        self.config.last_updated = time.time()
        self.updated.set()


    def update_status_info(self):
        queue = self.download_queue_info()
        slot_info = self.download_slot_info()
        self.config.update_status(queue, slot_info)


    def update(self):
        self.update_status_info()
        self.log_update()
        if not self.updated.isSet(): self.update_client_config()

        # TODO only mark status updated if something actually changed
        self.status_updated.set()


    # Commands
    def process_command_queue(self):
        while len(self.command_queue):
            try:
                self.send_command(self.command_queue.pop())
                self.config.last_update_attempt = 0 # Force an update
            except:
                if debug: traceback.print_exc()
                self.failed_commands += 1

        self.commands_flushed.notify()


    # State functions
    def offline_state(self):
        if self.last_connection + self.connect_rate < time.time():
            try:
                self.last_connection = time.time()
                self.connecting.set()
                self.connect_fail_reason = self.connect()
                if self.com.is_connected(): self.online.set()
                else: self.online.clear()
            finally:
                self.connecting.clear()


    def online_state(self):
        try:
            self.process_command_queue()

            if self.selected.isSet() and \
                    self.last_update_attempt + self.update_rate < time.time():
                self.last_update_attempt = time.time()
                self.update()

        except Exception, e:
            if debug: traceback.print_exc()
            print 'Update failed: %s' % e
            if debug: traceback.print_exc()

            # There was an error so reconnect
            self.online.clear()
            self.updated.clear()


    def run(self):
        self.lock.acquire()

        while not self.shutdown.isSet():
            try:
                if self.online.isSet(): self.online_state()
                else: self.offline_state()

                try:
                    self.lock.release()
                    if self.selected.isSet(): self.shutdown.wait(2)
                    else: self.shutdown.wait(5)

                finally:
                    self.lock.acquire()

            except:
                if debug: traceback.print_exc()

        self.lock.release()
